import "dart:convert";
import "package:collection/collection.dart";
import 'package:verttice_logic/verttice_logic.dart';
import 'package:verttice_ui/src/const/box_sizes.dart';
import "package:verttice_ui/src/const/icons.dart";
import 'package:verttice_ui/src/state/verttice_ui_state.dart';
import "package:verttice_ui/src/widgets/form_factory/form_factory.dart";
import "package:verttice_ui/src/widgets/form_factory/form_fields.dart";
import 'package:verttice_ui/src/widgets/tables/filter_view.dart';
import "package:verttice_ui/src/widgets/tables/table_utils.dart";
import "package:verttice_ui/src/widgets/themed_toast.dart";
import "package:data_table_2/data_table_2.dart";
import "package:flutter/material.dart";
import "package:flutter/scheduler.dart";
import "package:flutter/services.dart";
import "package:provider/provider.dart";

class TableContent extends StatefulWidget {
  final bool copyRow;
  final List<FormItem> columns;
  final List<Map<String, dynamic>> data;
  final int currentPage;
  final int pageLength;
  final void Function(
    Map<String, dynamic>? rowData,
    int rowIndex,
  ) handleSelectedRow;
  final void Function(List<Map<String, dynamic>>)? selectedItemListCallback;
  final void Function(int) filteredDataLength;
  final String search;
  final String clearAllFilters;
  final Function(List<FilterField>)? onFilterChanged;
  final Function(List<Map<String, dynamic>>) filteredDataCallback;

  const TableContent({
    Key? key,
    required this.copyRow,
    required this.columns,
    required this.data,
    required this.currentPage,
    required this.pageLength,
    required this.handleSelectedRow,
    required this.search,
    required this.filteredDataLength,
    required this.clearAllFilters,
    required this.filteredDataCallback,
    this.selectedItemListCallback,
    this.onFilterChanged,
  }) : super(key: key);

  @override
  _TableContentState createState() => _TableContentState();
}

class _TableContentState extends State<TableContent> {
  late List<Map<String, dynamic>> _data;
  late List<Map<String, dynamic>> _filteredData;
  final List<Map<String, dynamic>> _selctedRows = [];
  late final List<DataColumn2> _tableColumns = [];
  late final List<FormItem> _columns = [];
  final Map<String, Map<dynamic, FilterField>> _filters = {};
  late double _tableWidth;
  late String _search;
  late int _currentPage;
  late int _pageLength;
  late String _clearAllFilters;
  String? _selectedRow;
  int _currentSortColumn = 0;
  bool _isAscending = true;

  @override
  void initState() {
    super.initState();
    _search = widget.search;
    _data = widget.data;
    _filteredData = _data;
    _currentPage = widget.currentPage;
    _pageLength = widget.pageLength;
    _clearAllFilters = widget.clearAllFilters;
    if (widget.copyRow) {
      _columns.add(FormItem(field: NoneFF()));
    }
    if (widget.selectedItemListCallback != null) {
      _columns.add(FormItem(field: NoneFF()));
    }

    _columns.addAll(widget.columns);
    _tableWidth =
        _columns.map((e) => e.field.tableColumnWidth + 15).toList().sum;
    _buildColumns();
  }

  @override
  void didUpdateWidget(covariant TableContent oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (_clearAllFilters != widget.clearAllFilters) {
      _clearAllFilters = widget.clearAllFilters;
      _filteredData = _data;
      _filters.clear();
      _tableColumns.clear();
      _buildColumns();
    }
    if (_currentPage != widget.currentPage ||
        _pageLength != widget.pageLength) {
      _currentPage = widget.currentPage;
      _pageLength = widget.pageLength;
    }
    if (_search != widget.search && widget.onFilterChanged == null) {
      _filteredData = _filterResult(widget.search, _filters, _data);
      widget.filteredDataLength(_filteredData.length);
      _search = widget.search;
    }
    if (!const DeepCollectionEquality().equals(_data, widget.data)) {
      _data = widget.data;
      if (widget.onFilterChanged == null) {
        _filteredData = _filterResult(_search, _filters, _data);
        widget.filteredDataLength(_filteredData.length);
      } else {
        _filteredData = _data;
      }
    }
  }

  @override
  void setState(VoidCallback fn) {
    if (mounted) {
      SchedulerBinding.instance?.addPostFrameCallback((_) {
        super.setState(fn);
      });
    }
  }

  void _buildColumns() {
    for (var column in _columns) {
      if (_columns.indexOf(column) == 0 && widget.copyRow) {
        _tableColumns.add(const DataColumn2(
          size: ColumnSize.S,
          label: SelectableText(
            "Copy",
          ),
        ));
      } else if (widget.selectedItemListCallback != null &&
          ((!widget.copyRow && _columns.indexOf(column) == 0) ||
              _columns.indexOf(column) == 1)) {
        _tableColumns.add(const DataColumn2(
          size: ColumnSize.S,
          label: SelectableText(
            "Select",
          ),
        ));
      } else if (column.field.showOnTable &&
          column.field.displayName != "" &&
          column.field.name != "") {
        _tableColumns.add(_builDataColumn(column: column));
      }
    }
  }

  List<Map<String, dynamic>> _searchResult(
    String search,
    List<Map<String, dynamic>> data,
  ) {
    return search == ""
        ? data
        : data
            .where(
              (element) => json.encode(element).toLowerCase().contains(
                    search.toLowerCase(),
                  ),
            )
            .toList();
  }

  DataColumn2 _builDataColumn({required FormItem column}) {
    return DataColumn2(
      size: column.field.tableColumnWidth <= 100
          ? ColumnSize.S
          : column.field.tableColumnWidth <= 150
              ? ColumnSize.M
              : ColumnSize.L,
      onSort: (int columnIndex, _) {
        setState(() {
          _currentSortColumn = columnIndex;
          if (_isAscending == true) {
            _isAscending = false;
          } else {
            _isAscending = true;
          }
        });

        _handleSortData(
          _columns[columnIndex],
          _columns[columnIndex].field.name,
          !_isAscending,
        );
      },
      label: Column(
        mainAxisAlignment: MainAxisAlignment.start,
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          sizedBoxH4,
          SizedBox(
            height: 30,
            child: SelectableText(
              column.field.displayName,
              style: const TextStyle(
                overflow: TextOverflow.fade,
              ),
            ),
          ),
          SizedBox(
            height: 30,
            child: TextButton.icon(
              icon: Tooltip(
                message: "Apply filter",
                child: ThemedIcons.filter(
                  iconStyle: IconStyle(
                    size: 17,
                  ),
                ),
              ),
              label: Builder(builder: (context) {
                final labletext = _filters[column.field.name]
                        ?.values
                        .map(
                          (e) => visualDataProcessor(
                            column,
                            {
                              column.field.name: e.value,
                            },
                          ).value,
                        )
                        .toList()
                        .join(", ") ??
                    "";
                return Tooltip(
                  message: labletext,
                  child: Text(
                    labletext,
                  ),
                );
              }),
              onPressed: () => _filterView(column: column),
            ),
          ),
          sizedBoxH2,
        ],
      ),
    );
  }

  Future<void> _filterView({
    required FormItem column,
  }) async {
    await showDialog(
      context: context,
      useSafeArea: true,
      builder: (BuildContext context) {
        return SimpleDialog(
          insetPadding: const EdgeInsets.all(6),
          contentPadding: const EdgeInsets.all(6),
          children: [
            SizedBox(
              height: 450,
              width: 400,
              child: Scaffold(
                body: Center(
                  child: FilterView(
                    column: column,
                    filter: _filters[column.field.name],
                    clearFilterCallback: () {
                      setState(() {
                        _filters.remove(column.field.name);

                        int index = _columns.indexOf(column);
                        _tableColumns.replaceRange(
                          index,
                          index + 1,
                          [_builDataColumn(column: column)],
                        );
                        if (widget.onFilterChanged != null) {
                          List<FilterField> filters = [];
                          _filters.forEach((key, value) {
                            filters.add(
                              FilterField(
                                field: value.values.first.field,
                                filterType: value.values.first.filterType,
                                value: value.values.length == 1
                                    ? value.values.first.value
                                    : value.values.map((e) => e.value).toList(),
                              ),
                            );
                          });
                          widget.onFilterChanged!(filters);
                        } else {
                          _filteredData = _filteredData.where((element) {
                            bool include = true;
                            for (var entry in _filters.entries) {
                              for (var filter in entry.value.entries) {
                                final applFil = applayTableFilter(
                                  filter: filter.value,
                                  data: element[entry.key],
                                );
                                if (!applFil) {
                                  include = false;
                                }
                              }
                            }
                            return include;
                          }).toList();
                          widget.filteredDataLength(_filteredData.length);
                          widget.filteredDataCallback(_filteredData);
                        }
                      });
                    },
                    filterCallback: (Map<dynamic, FilterField> filter) {
                      setState(() {
                        _filters[column.field.name] = filter;

                        int index = _columns.indexOf(column);
                        _tableColumns.replaceRange(
                          index,
                          index + 1,
                          [_builDataColumn(column: column)],
                        );
                        if (widget.onFilterChanged != null) {
                          List<FilterField> filters = [];
                          _filters.forEach((key, value) {
                            filters.add(
                              FilterField(
                                field: value.values.first.field,
                                filterType: value.values.first.filterType,
                                value: value.values.length == 1
                                    ? value.values.first.value
                                    : value.values.map((e) => e.value).toList(),
                              ),
                            );
                          });
                          widget.onFilterChanged!(filters);
                        } else {
                          _filteredData = _filteredData.where((element) {
                            bool include = true;
                            for (var entry in _filters.entries) {
                              for (var filter in entry.value.entries) {
                                final applFil = applayTableFilter(
                                  filter: filter.value,
                                  data: element[entry.key],
                                );
                                if (!applFil) {
                                  include = false;
                                }
                              }
                            }
                            return include;
                          }).toList();
                          widget.filteredDataCallback(_filteredData);
                          widget.filteredDataLength(_filteredData.length);
                        }
                      });
                    },
                  ),
                ),
              ),
            )
          ],
        );
      },
    );
  }

  List<Map<String, dynamic>> _filterResult(
    String search,
    Map<String, Map<dynamic, FilterField>> filters,
    List<Map<String, dynamic>> data,
  ) {
    final result = _searchResult(search, _data).where((element) {
      bool include = true;
      for (var entry in _filters.entries) {
        for (var filter in entry.value.entries) {
          final applFil = applayTableFilter(
            filter: filter.value,
            data: element[entry.key],
          );
          if (!applFil) {
            include = false;
          }
        }
      }
      return include;
    }).toList();

    return result;
  }

  void _handleSortData(
    FormItem column,
    String name,
    bool desc,
  ) {
    _filteredData.sort((a, b) {
      dynamic aS = a[name] ?? "";
      dynamic bS = b[name] ?? "";
      if (column.field.pathToField != null &&
          column.field.pathToField!.isNotEmpty) {
        aS = fieldFromPathToField(
          column.field.pathToField!,
          a.cast<String, dynamic>(),
        );
        bS = fieldFromPathToField(
          column.field.pathToField!,
          b.cast<String, dynamic>(),
        );
      }

      if (aS == null || aS == "") {
        if (bS == null || bS == "") {
          return 0;
        } else {
          return 1;
        }
      } else {
        if (bS == null || bS == "") {
          return -1;
        }
      }

      var r = desc
          ? bS.toString().compareTo(aS.toString())
          : aS.toString().compareTo(bS.toString());
      if (r != 0) return r;
      return desc
          ? bS.toString().compareTo(aS.toString())
          : aS.toString().compareTo(bS.toString());
    });
  }

  DataRow _dataRowBuilder(
    String? selectedRow,
    int rowIndex,
    Map<String, dynamic> rowData,
    bool isDarkTheme,
  ) {
    final List<DataCell> currenRow = [];

    for (var column in _columns) {
      if (_columns.indexOf(column) == 0 && widget.copyRow) {
        currenRow.add(
          DataCell(
            _copyRow(rowData),
          ),
        );
      } else if (widget.selectedItemListCallback != null &&
          ((!widget.copyRow && _columns.indexOf(column) == 0) ||
              _columns.indexOf(column) == 1)) {
        final val = _selctedRows.firstWhereOrNull(
                (element) => const MapEquality().equals(element, rowData)) !=
            null;
        currenRow.add(
          DataCell(
            Tooltip(
              message: val ? "Uncheck" : "Check",
              child: Checkbox(
                onChanged: (selcted) {
                  setState(() {
                    if (selcted ?? false) {
                      _selctedRows.add(rowData);
                    } else {
                      _selctedRows.remove(rowData);
                    }

                    widget.selectedItemListCallback!(_selctedRows);
                  });
                },
                value: val,
              ),
            ),
          ),
        );
      } else if (column.field.showOnTable &&
          column.field.displayName != "" &&
          column.field.name != "") {
        final visualData = visualDataProcessor(column, rowData);
        final text = visualData.value;
        final widget = visualData.widget;
        currenRow.add(
          DataCell(InkWell(
            onTap: () {
              Clipboard.setData(
                ClipboardData(
                  text: text,
                ),
              );
              ThemedToast.info(
                context: context,
                message: "Copied to clipboard",
              );
            },
            child: widget,
          )),
        );
      }
    }

    final stringrow = rowData['createdAt']?.toString();

    return DataRow(
      onSelectChanged: (bool? selected) {
        widget.handleSelectedRow(rowData, rowIndex);

        setState(() {
          _selectedRow = stringrow;
        });
      },
      color: selectedRow == stringrow && stringrow != null
          ? MaterialStateProperty.all(
              isDarkTheme ? Colors.black : Colors.grey[400],
            )
          : ((rowIndex % 2) == 0)
              ? MaterialStateProperty.all(
                  isDarkTheme ? Colors.grey[600] : Colors.grey[50],
                )
              : MaterialStateProperty.all(
                  isDarkTheme ? Colors.grey[800] : Colors.grey[200],
                ),
      cells: currenRow,
    );
  }

  Widget _copyRow(Map<String, dynamic> rowData) {
    return Tooltip(
      message: "Copy row",
      child: IconButton(
        onPressed: () async {
          await Clipboard.setData(
            ClipboardData(
              text: widget.columns
                  .map((e) =>
                      "${e.field.displayName}: ${visualDataProcessor(e, rowData).value}")
                  .toList()
                  .join(',\n'),
            ),
          );
          ThemedToast.info(
            context: context,
            message: "Row copied to clipboard",
          );
        },
        icon: ThemedIcons.copy(),
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    final dataLength = _filteredData.length;

    final sublist = _filteredData.sublist(
        (_currentPage * _pageLength) > dataLength
            ? dataLength
            : (_currentPage * _pageLength),
        ((_currentPage + 1) * _pageLength) > dataLength
            ? dataLength
            : ((_currentPage + 1) * _pageLength));

    return Consumer<VertticeUIState>(builder: (context, value, _child) {
      return DataTable2(
        showBottomBorder: true,
        scrollController: ScrollController(),
        minWidth: _tableWidth,
        sortColumnIndex: _currentSortColumn,
        sortAscending: _isAscending,
        headingRowHeight: 68,
        dataRowHeight: 55,
        columnSpacing: 1,
        columns: _tableColumns,
        rows: sublist.mapIndexed((rowIndex, rowData) {
          return _dataRowBuilder(
            _selectedRow,
            rowIndex,
            rowData,
            value.isDarkTheme,
          );
        }).toList(),
        showCheckboxColumn: false,
      );
    });
  }
}
