import "dart:math";
import "package:collection/collection.dart";
import "package:verttice_ui/src/widgets/form_factory/form_factory.dart";
import "package:verttice_ui/src/widgets/tables/table_frame.dart";
import "package:verttice_ui/src/widgets/themed_divider.dart";
import "package:flutter/material.dart";
import "package:flutter/scheduler.dart";
import 'package:verttice_logic/verttice_logic.dart';

class TableData extends StatefulWidget {
  final List<Map<String, dynamic>> data;
  final List<FormItem> columns;
  final String title;
  final void Function(Map<String, dynamic>? rowData)? rowPressCallback;
  final Widget? customTopBarWidget;
  final double? customTopBarPresntage;
  final bool showTopBar;
  final List<int> rowsPerPageOptions;
  final bool copyRow;
  final List<Widget>? actions;
  final void Function(List<Map<String, dynamic>>)? selectedItemListCallback;
  final void Function(Map<String, dynamic> data)? topBarCheckDataCallback;
  final double? actionPrecentageWidth;
  final VoidCallback? onEndReached;
  final Function(List<FilterField>)? onFilterChanged;
  final Function(String)? onSearchChanged;
  const TableData(
      {Key? key,
      required this.data,
      required this.columns,
      this.title = "Table",
      this.rowPressCallback,
      this.customTopBarWidget,
      this.customTopBarPresntage,
      this.showTopBar = true,
      this.topBarCheckDataCallback,
      this.selectedItemListCallback,
      this.rowsPerPageOptions = const [20, 50, 100],
      this.copyRow = true,
      this.actions,
      this.actionPrecentageWidth,
      this.onEndReached,
      this.onFilterChanged,
      this.onSearchChanged})
      : super(key: key);

  @override
  _TableDataState createState() => _TableDataState();
}

class _TableDataState extends State<TableData> {
  final double _seperator = 2.0;
  late List<Map<String, dynamic>> _tableData;
  late double _topHeightPresntage;

  @override
  void initState() {
    super.initState();
    _tableData = widget.data;
    _topHeightPresntage =
        (widget.customTopBarWidget != null) || widget.showTopBar
            ? min(widget.customTopBarPresntage ?? 0.4, 0.4)
            : 0.0;
  }

  @override
  void setState(VoidCallback fn) {
    if (mounted) {
      SchedulerBinding.instance?.addPostFrameCallback((_) {
        super.setState(fn);
      });
    }
  }

  @override
  void didUpdateWidget(covariant TableData oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (!const DeepCollectionEquality().equals(_tableData, widget.data)) {
      _tableData = widget.data;
    }
  }

  @override
  Widget build(BuildContext context) {
    return LayoutBuilder(builder: (context, constraints) {
      final double topPanelHeight =
          (constraints.maxHeight * _topHeightPresntage);
      return Stack(
        children: [
          SizedBox(
            height: constraints.maxHeight,
            width: constraints.maxWidth,
          ),
          AnimatedPositioned(
              duration: const Duration(milliseconds: 200),
              top: 0,
              child: (widget.customTopBarWidget != null) || widget.showTopBar
                  ? SizedBox(
                      height: topPanelHeight,
                      width: constraints.maxWidth,
                      child: widget.customTopBarWidget,
                    )
                  : const SizedBox(
                      height: 0,
                      width: 0,
                    )),
          AnimatedPositioned(
            duration: const Duration(milliseconds: 200),
            top: topPanelHeight,
            height: _seperator,
            width: constraints.maxWidth,
            child: MouseRegion(
                cursor: SystemMouseCursors.resizeRow,
                child: GestureDetector(
                  onPanUpdate: (DragUpdateDetails dragDetail) {
                    setState(() {
                      _topHeightPresntage =
                          dragDetail.globalPosition.dy / constraints.maxHeight;
                    });
                  },
                  child: ThemedDivider(
                    height: _seperator,
                    width: constraints.maxWidth,
                  ),
                )),
          ),
          AnimatedPositioned(
            duration: const Duration(milliseconds: 300),
            top: topPanelHeight + _seperator,
            height: constraints.maxHeight - (topPanelHeight + _seperator),
            width: constraints.maxWidth,
            child: TableFrame(
              data: _tableData,
              onEndReached: () {
                if (widget.onEndReached != null) {
                  widget.onEndReached!();
                }
              },
              onSearchChanged: widget.onSearchChanged,
              onFilterChanged: widget.onFilterChanged,
              actionPrecentageWidth: widget.actionPrecentageWidth,
              frameWidth: constraints.maxWidth,
              frameHeight:
                  constraints.maxHeight - (topPanelHeight + _seperator),
              actions: widget.actions,
              copyRow: widget.copyRow,
              selectedItemListCallback: widget.selectedItemListCallback,
              columns: widget.columns,
              rowsPerPageOptions: widget.rowsPerPageOptions,
              title: widget.title,
              handleSelectedRow: (Map<String, dynamic>? rowData, int rowIndex) {
                if (widget.rowPressCallback != null) {
                  widget.rowPressCallback!(rowData);
                }
              },
            ),
          )
        ],
      );
    });
  }
}
